#include <Wire.h>
#include <TinyGPS++.h>
#include <SoftwareSerial.h>
#include <LiquidCrystal_I2C.h>
#include <Keypad.h>
#include <Servo.h>
#include <EEPROM.h>

// Codes
int codeCounter = 0;
char code1[7] = "111111";
char code2[7] = "222222";
char code3[7] = "333333";
char initCode[7] = "000000";
char masterCode[7] = "999999";
char codeEntered[7] = "";

// Locations
static const double LOCATION_1_LAT = 36.101422, LOCATION_1_LON = -112.118490; // Grand Canyon
static const double LOCATION_2_LAT = 36.101422, LOCATION_2_LON = -112.118490; // Grand Canyon
static const double LOCATION_3_LAT = 36.101422, LOCATION_3_LON = -112.118490; // Grand Canyon

static const double margin = 0.1; // distance of margin of error in miles

// Servos
Servo servo1;
Servo servo2;
Servo servo3;
static const int ServoPin1 = 10, ServoPin2 = 11, ServoPin3 = 12;

// LCD
LiquidCrystal_I2C lcd(0x27, 16, 2); // set the LCD address to 0x27 for a 16 chars and 2 line display

// Keypad
const byte ROWS = 4; //four rows
const byte COLS = 4; //three columns
char keys[ROWS][COLS] = {
  {'1', '2', '3', 'A'},
  {'4', '5', '6', 'B'},
  {'7', '8', '9', 'C'},
  {'*', '0', '#', 'D'}
};
byte rowPins[ROWS] = {0, 1, 2, 3}; //connect to the row pinouts of the keypad
//byte colPins[COLS] = {13, 5, 6, 7}; //connect to the column pinouts of the keypad
byte colPins[COLS] = {4, 5, 6, 7}; //connect to the column pinouts of the keypad
Keypad keypad = Keypad( makeKeymap(keys), rowPins, colPins, ROWS, COLS );

// Key switches
static const int KeyPin1 = 15, KeyPin2 = 14;

// GPS
static const int RXPin = 9, TXPin = 8;
static const uint32_t GPSBaud = 9600;

// The TinyGPS++ object
TinyGPSPlus gps;

// The serial connection to the GPS device
SoftwareSerial ss(RXPin, TXPin);

// The state of the side doors
boolean door1IsOpen = false, door2IsOpen = false, door3IsOpen = false;
boolean door1IsOpenNow = false, door2IsOpenNow = false, door3IsOpenNow = false;
int door1Open = 180, door1Close = 0;
int door2Open = 0, door2Close = 185;
int door3Open = 90, door3Close = 180;

void setup()
{
  lcd.init();                      // initialize the lcd
  lcd.backlight();

  // Init servos
  //servo1.attach(ServoPin1); 
  //servo2.attach(ServoPin2); 
  //servo3.attach(ServoPin3); 
  
  // Init previous states
  door1IsOpen = (boolean)(EEPROM.read(0));
  door2IsOpen = (boolean)(EEPROM.read(1));
  door3IsOpen = (boolean)(EEPROM.read(2));
  door1IsOpenNow = (boolean)(EEPROM.read(3));
  door2IsOpenNow = (boolean)(EEPROM.read(4));
  door3IsOpenNow = (boolean)(EEPROM.read(5));
  startupDoors();

  // Key switches
  pinMode(KeyPin1, INPUT_PULLUP); // key switch 1
  pinMode(KeyPin2, INPUT_PULLUP); // key switch 2

  // Init Readout
  initScreen();
}

void loop()
{
  char key = keypad.getKey();

  if (key != NO_KEY) {
    lcd.print(key);
    
    // Store the new key in the current entry string
    codeEntered[codeCounter] = key;
    codeCounter++;
    
    // If we are at 6 characters, see if it is a match
    if (codeCounter == 6)
    {
      int match = keyIsMatch();
      
      if (match > 0)
      {
        // If it is 1 or 2, check the location, open the door
        if (match == 1 || match == 2)
        {
          if ((match == 1 && door1IsOpenNow)
          || (match == 2 && door2IsOpenNow))
          {
            closeDoor(match);
          }
          else if ((match == 1 && door1IsOpen && !door1IsOpenNow)
          || (match == 2 && door2IsOpen && !door2IsOpenNow))
          {
            openDoor(match);
          }
          else if (checkLocation(match))
          {
            openDoor(match);
          } 
        }
        else if (match == 3)// last code/location
        {
          if (door3IsOpenNow)
          {
            closeDoor(match);
          }
          else if (door3IsOpen && !door3IsOpenNow)
          {
            openDoor(match);
          }
          else if (digitalRead(KeyPin1) == LOW && digitalRead(KeyPin2) == LOW)
          {
            if (door1IsOpen && door2IsOpen)
            {
              if (checkLocation(match))
              {
                // Both keys needs to be turned, both doors open, and in the right location
                openDoor(match); 
              }
            }
            else
            {
              lcd.clear();
              lcd.print("Cheater!");
              
              delay(2000);
               
              lcd.clear();
              lcd.print("Open both");
              lcd.setCursor(0,1);
              lcd.print("doors first"); 
              
              delay(2000);
            }
          }
          else
          {
            lcd.clear();
            lcd.print("Keys needed");
          }          
        }
        else if (match == 4)
        {
          // lock box
          initBox();
        }
        else if (match == 5)
        {
          // unlock box
          masterUnlock();
        }
      }
      else
      {
         lcd.clear();
         lcd.print("Invalid code");
      }
      
      delay(3000);
      initScreen();
      
      codeCounter = 0;
    }
   
  }

  /*int state = digitalRead(KEY1);
  if (state == HIGH)
  {
    lcd.clear();
    lcd.print("Turned");
  }
  else
  {
    lcd.clear();
    lcd.print("Off");
  }*/
}

void startupDoors()
{
  lcd.clear();
  lcd.print("Starting...");
  
  servo1.attach(ServoPin1);
  if (door1IsOpen)
    servo1.write(door1Open);
  else
    servo1.write(door1Close);
  
  delay(500);
  servo1.detach();
  delay(500);
  
  servo2.attach(ServoPin2);
  if (door2IsOpen)
    servo2.write(door2Open);
  else
    servo2.write(door2Close);
  
  delay(500);
  servo2.detach();
  delay(500);
  
  servo3.attach(ServoPin3);
  if (door3IsOpen)
    servo3.write(door3Open);
  else
    servo3.write(door3Close);
  
  delay(500);
  servo3.detach();
  delay(500);
}

void initBox()
{
 // Set the starting servo positions
  lcd.clear();
  lcd.print("Initialize");
  
  servo1.attach(ServoPin1); 
  servo1.write(door1Close);
  delay(500);
  servo1.detach(); 
  
  delay(500);
  
  servo2.attach(ServoPin2); 
  servo2.write(door2Close);
  delay(500);
  servo2.detach(); 
  
  delay(500);
  
  servo3.attach(ServoPin3); 
  servo3.write(door3Close);
  delay(500);
  servo3.detach(); 
  
  door1IsOpen = false;
  door2IsOpen = false;
  door3IsOpen = false;
  door1IsOpenNow = false;
  door2IsOpenNow = false;
  door3IsOpenNow = false;
  EEPROM.write(0, door1IsOpen);
  EEPROM.write(1, door2IsOpen);
  EEPROM.write(2, door3IsOpen);
  EEPROM.write(3, door1IsOpenNow);
  EEPROM.write(4, door2IsOpenNow);
  EEPROM.write(5, door3IsOpenNow);
  
  lcd.clear();
}

void masterUnlock()
{
  // Open all doors
  lcd.clear();
  lcd.print("Master Unlock");
  
  servo1.attach(ServoPin1); 
  servo1.write(door1Open);
  delay(500);
  servo1.detach(); 
  
  delay(500);
  
  servo2.attach(ServoPin2); 
  servo2.write(door2Open);
  delay(500);
  servo2.detach(); 
  
  delay(500);
  
  servo3.attach(ServoPin3); 
  servo3.write(door3Open);
  delay(500);
  servo3.detach(); 
  
  door1IsOpen = true;
  door2IsOpen = true;
  door3IsOpen = true;
  door1IsOpenNow = true;
  door2IsOpenNow = true;
  door3IsOpenNow = true;
  EEPROM.write(0, door1IsOpen);
  EEPROM.write(1, door2IsOpen);
  EEPROM.write(2, door3IsOpen);
  EEPROM.write(3, door1IsOpenNow);
  EEPROM.write(4, door2IsOpenNow);
  EEPROM.write(5, door3IsOpenNow);
  
  lcd.clear();
}

int keyIsMatch()
{
  if (strcmp(codeEntered,code1) == 0)
    return 1;
  if (strcmp(codeEntered,code2) == 0)
    return 2;
  if (strcmp(codeEntered,code3) == 0)
    return 3;
  if (strcmp(codeEntered,initCode) == 0)
    return 4;
  if (strcmp(codeEntered,masterCode) == 0)
    return 5;
    
  return 0;
}

bool checkLocation(int locationNum)
{
  // Check the location against the desired number.
  double latitude, longitude;
  
  if (locationNum == 1)
  {
    latitude = LOCATION_1_LAT;
    longitude = LOCATION_1_LON;
  }
  else if (locationNum == 2)
  {
    latitude = LOCATION_2_LAT;
    longitude = LOCATION_2_LON;
  }
  else // 3
  {
    latitude = LOCATION_3_LAT;
    longitude = LOCATION_3_LON;
  }
  
  bool foundLocation = false;
  int timeToCheck = 60;
  
  ss.begin(GPSBaud);
  
  // Get the current location
  // Dispatch incoming characters
  do
  {
    if (gps.location.isValid())
    {
      foundLocation = true;
      
      double distanceToLocation =
          TinyGPSPlus::distanceBetween(
            gps.location.lat(),
            gps.location.lng(),
            latitude, 
            longitude);
        distanceToLocation = (distanceToLocation/1000)*0.6; // Convert to miles
        
        lcd.clear();
        if (distanceToLocation < margin)
        {
          lcd.print("Congrats!");
          delay(2000);
          
          ss.end();
          return true;
        }
        else
        {
          lcd.print(distanceToLocation);
          lcd.print("mi");
          lcd.setCursor(0,1);
          lcd.print("from destination");
          timeToCheck -= 3; // Speed up
        }
    }
    
    if (gps.charsProcessed() < 10)
    {
     lcd.clear();
     lcd.print("Checking"); 
     lcd.setCursor(0,1);
     lcd.print("Location..."); 
    }
  
    timeToCheck--;
    smartDelay(1000);
  }
  while (timeToCheck > 0);
  
  lcd.clear();
  
  if (!foundLocation)
  {
    lcd.print("No Signal");
    lcd.setCursor(0,1);
    lcd.print("Go Outside!"); 
  }
  else
  {
    lcd.print("Resetting...");
  }
  
  ss.end();
  return false; 
}

// This custom version of delay() ensures that the gps object
// is being "fed".
static void smartDelay(unsigned long ms)
{
  unsigned long start = millis();
  do 
  {
    while (ss.available())
      gps.encode(ss.read());
  } while (millis() - start < ms);
}

void openDoor(int doorNum)
{
  // Open the servo for the matching door
  
  if (doorNum == 1)
  {
    servo1.attach(ServoPin1); 
    servo1.write(door1Open);
    delay(500);
    servo1.detach();
    
    door1IsOpen = true;
    door1IsOpenNow = true;
    EEPROM.write(0, door1IsOpen);
    EEPROM.write(3, door1IsOpenNow);
  }
  else if (doorNum == 2)
  {
    servo2.attach(ServoPin2); 
    servo2.write(door2Open);
    delay(500);
    servo2.detach();
    
    door2IsOpen = true;
    door2IsOpenNow = true;
    EEPROM.write(1, door2IsOpen);
    EEPROM.write(4, door2IsOpenNow);
  }
  else
  {
    servo3.attach(ServoPin3); 
    servo3.write(door3Open);
    delay(500);
    servo3.detach();
    
    door3IsOpen = true;
    door3IsOpenNow = true;
    EEPROM.write(2, door3IsOpen);
    EEPROM.write(5, door3IsOpenNow);
  }
  
  lcd.clear();
  lcd.print("Door ");
  lcd.print(doorNum);
  lcd.print(" opened");
}

void closeDoor(int doorNum)
{
  // Open the servo for the matching door
  
  if (doorNum == 1)
  {
    servo1.attach(ServoPin1); 
    servo1.write(door1Close);
    delay(500);
    servo1.detach();
    
    door1IsOpenNow = false;
    EEPROM.write(3, door1IsOpenNow);
  }
  else if (doorNum == 2)
  {
    servo2.attach(ServoPin2); 
    servo2.write(door2Close);
    delay(500);
    servo2.detach();
    
    door2IsOpenNow = false;
    EEPROM.write(4, door2IsOpenNow);
  }
  else
  {
    servo3.attach(ServoPin3); 
    servo3.write(door3Close);
    delay(500);
    servo3.detach();
    
    door3IsOpenNow = false;
    EEPROM.write(5, door3IsOpenNow);
  }
  
  lcd.clear();
  lcd.print("Door ");
  lcd.print(doorNum);
  lcd.print(" closed");
}

void initScreen()
{
  lcd.clear();
  lcd.print("Enter Code");
  lcd.setCursor(0,1);
}
